from httpx import AsyncClient

from chatushka import CommandsMatcher, Message, Telegram
from chatushka.bot.settings import BOBUK_JOKES_URL, get_settings

settings = get_settings()
jokes_matcher = CommandsMatcher(
    prefixes=settings.command_prefixes,
    postfixes=settings.command_postfixes,
)


@jokes_matcher("jokes", "joke")
async def jokes_handler(
    api: Telegram,
    message: Message,
) -> None:
    async with AsyncClient() as client:  # type: AsyncClient
        response = await client.get(BOBUK_JOKES_URL)
        try:
            response.raise_for_status()
            joke = response.json()["content"]
        except Exception:  # noqa, pylint: disable=broad-except
            return
    await api.send_message(
        chat_id=message.chat.id,
        text=joke,
    )



from datetime import datetime
from logging import getLogger
from typing import Any

from httpx import AsyncClient, Response
from pydantic import ValidationError

from chatushka.core.models import (
    ChatMemberAdministrator,
    ChatMemberOwner,
    ChatMemberStatuses,
    ChatPermissions,
    Message,
    Update,
    User,
)

logger = getLogger()


class Telegram:
    def __init__(
        self,
        token: str,
    ) -> None:
        self.token = token

    @property
    def _base_api_url(self) -> str:
        return f"https://api.telegram.org/bot{self.token}"

    def _api_method_url(
        self,
        method: str,
    ) -> str:
        return f"{self._base_api_url}/{method}"

    @staticmethod
    def check_api_response(
        response: Response,
    ) -> dict[str, Any]:
        data: dict[str, Any] = response.json()
        is_ok: bool = data.get("ok", False)
        if not is_ok:
            logger.warning(response.text)
            raise ValueError(f"Telegram response error: {response.text}\n{data}")

        result: dict[str, Any] = data["result"]
        return result

    async def _call_api(
        self,
        method: str,
        timeout: int = 10,
        **kwargs: Any,
    ) -> list[dict[str, Any]] | dict[str, Any]:
        url = self._api_method_url(method)
        async with AsyncClient() as client:  # type: AsyncClient
            response = await client.post(url, timeout=timeout * 2, data=kwargs)
        return self.check_api_response(response)

    async def get_me(
        self,
    ) -> User:
        result = await self._call_api("getme")
        return User(**result)  # type: ignore

    async def get_updates(
        self,
        timeout: int,
        offset: None | int = None,
        polling: bool = False,
    ) -> tuple[list[Update], int]:
        params = {}
        if offset:
            params["offset"] = offset
        if polling:
            params["timeout"] = timeout
        results = await self._call_api(
            "getupdates",
            timeout=timeout,
            **params,
        )
        updates_list = []
        latest_update_id: int | None = offset
        for result in results:
            if not latest_update_id or (latest_update_id < result["update_id"]):  # type: ignore
                latest_update_id = result["update_id"]  # type: ignore
            try:
                update = Update(**result)  # type: ignore
            except ValidationError as err:
                logger.debug(err)
                continue
            updates_list.append(update)
        return updates_list, latest_update_id  # type: ignore

    async def send_message(
        self,
        chat_id: int,
        text: str,
        reply_to_message_id: int | None = None,
        parse_mode: str = "html",
        disable_web_page_preview: bool = False,
    ) -> Message:
        result = await self._call_api(
            "sendmessage",
            chat_id=chat_id,
            text=text,
            reply_to_message_id=reply_to_message_id,
            parse_mode=parse_mode,
            disable_web_page_preview=disable_web_page_preview,
        )
        return Message(**result)  # type: ignore

    async def restrict_chat_member(
        self,
        chat_id: int,
        user_id: int,
        permissions: ChatPermissions,
        until_date: datetime,
    ) -> bool:
        result = await self._call_api(
            "restrictChatMember",
            chat_id=chat_id,
            user_id=user_id,
            permissions=permissions.json(),
            until_date=int(until_date.timestamp()),
        )
        return result  # noqa, type: ignore

    async def get_chat_administrators(
        self,
        chat_id: int,
    ) -> list[ChatMemberAdministrator | ChatMemberOwner]:
        results = await self._call_api(
            "getChatAdministrators",
            chat_id=chat_id,
        )
        admins = []
        for result in results:
            status = result["status"]
            if status == ChatMemberStatuses.CREATOR:
                admins.append(ChatMemberOwner(**result))
            if status == ChatMemberStatuses.ADMINISTRATOR:
                admins.append(ChatMemberAdministrator(**result))
        return admins

    async def pin_chat_message(
        self,
        chat_id: int,
        message_id: int,
        disable_notification: bool = True,
    ) -> bool:
        result = await self._call_api(
            "pinChatMessage",
            chat_id=chat_id,
            message_id=message_id,
            disable_notification=disable_notification,
        )
        return result  # noqa, type: ignore

    async def unpin_chat_message(
        self,
        chat_id: int,
        message_id: int,
    ) -> bool:
        result = await self._call_api(
            "unpinChatMessage",
            chat_id=chat_id,
            message_id=message_id,
        )
        return result  # noqa, type: ignore

    async def unpin_all_chat_messages(
        self,
        chat_id: int,
    ) -> bool:
        result = await self._call_api(
            "unpinAllChatMessages",
            chat_id=chat_id,
        )
        return result  # noqa, type: ignore



from random import choice

from chatushka import CommandsMatcher, Message, Telegram
from chatushka.bot.internal.data_dir import read_txt_from_data_dir
from chatushka.bot.settings import get_settings

settings = get_settings()
lukashenko_matcher = CommandsMatcher(
    prefixes=settings.command_prefixes,
    postfixes=settings.command_postfixes,
)


@lukashenko_matcher(
    "lukashenko",
    "luk",
    "lag",
    help_message="Quotes by Alexander G. Lukashenko",
)
async def lukashenko_handler(
    api: Telegram,
    message: Message,
) -> None:
    quotes = read_txt_from_data_dir("lukashenko.txt")
    await api.send_message(
        chat_id=message.chat.id,
        text=choice(quotes),
        reply_to_message_id=message.message_id,
    )



from functools import lru_cache
from pathlib import Path

from pydantic import BaseSettings

BOBUK_JOKES_URL = "https://jokesrv.rubedo.cloud/"
BOT_ROOT_DIR = Path(__file__).parent.resolve()
BOT_DATA_DIR = BOT_ROOT_DIR / "data"


class _Settings(BaseSettings):
    command_prefixes: str | tuple[str, ...] = ("/", "!")
    command_postfixes: str | tuple[str, ...] = "!"

    class Config:
        env_prefix = "BOT_"
        env_file = ".env"
        allow_mutation = False


@lru_cache
def get_settings() -> _Settings:
    return _Settings()



import signal
from asyncio import ensure_future, get_event_loop, sleep
from functools import partial
from logging import getLogger

from chatushka.__version__ import __URL__, __VERSION__
from chatushka.core.constants import HTTP_POOLING_DELAY, HTTP_POOLING_TIMEOUT
from chatushka.core.matchers import CommandsMatcher, EventsMatcher
from chatushka.core.models import EventTypes, Message
from chatushka.core.telegram import Telegram
from chatushka.core.utils import check_preconditions

logger = getLogger(__name__)


async def _message_handler(
    bot_instance: "Chatushka",
    message: Message,
    api: Telegram,
) -> None:
    await api.send_message(
        chat_id=message.chat.id,
        text=bot_instance.help_message_text,
        reply_to_message_id=message.message_id,
        parse_mode="markdown",
        disable_web_page_preview=True,
    )


class Chatushka(EventsMatcher):
    def __init__(
        self,
        token: str,
        title: str = None,
        debug: bool = False,
    ) -> None:
        super().__init__()

        self.title = title or self.__class__.__name__
        self.debug = debug
        self.api = Telegram(token)
        self.add_handler(EventTypes.STARTUP, check_preconditions, include_in_help=False)

        bot_commands_matcher = CommandsMatcher(prefixes=("!", "/"))
        bot_commands_matcher.add_handler(
            ("start", "help"),
            partial(_message_handler, self),
            help_message="This message!",
        )
        self.add_matcher(bot_commands_matcher)

    @property
    def help_message_text(self) -> str:
        output = f"*CHATUSHKA BOT {__VERSION__}*\n{__URL__}"
        for help_message in self.help_messages:
            output += f"\n\n*{', '.join(help_message.tokens)}*\n> {help_message.message}"
        return output

    # pylint: disable=too-many-nested-blocks
    async def _loop(self) -> None:
        offset: int | None = None
        while True:
            try:
                updates, latest_update_id = await self.api.get_updates(
                    timeout=HTTP_POOLING_TIMEOUT,
                    offset=offset,
                )
                if updates:
                    offset = latest_update_id + 1
            except Exception as err:  # noqa, pylint: disable=broad-except
                logger.error(err)
                await sleep(HTTP_POOLING_DELAY)
                continue
            if updates:
                for update in updates:
                    try:
                        for matcher in self.matchers:
                            matched_handlers = await matcher.match(self.api, update, should_call_matched=True)
                            if matched_handlers:
                                logger.debug(f"Matched {len(matched_handlers)} handlers")
                    except Exception as err:  # noqa, pylint: disable=broad-except
                        if self.debug:
                            raise
                        logger.error(err)
            await sleep(HTTP_POOLING_DELAY)

    async def _close(self) -> None:
        await self.call(self.api, EventTypes.SHUTDOWN)
        for matcher in self.matchers:
            if isinstance(matcher, EventsMatcher):
                await matcher.call(api=self.api, token=EventTypes.SHUTDOWN)

    async def serve(self) -> None:
        await self.call(self.api, EventTypes.STARTUP)
        loop = get_event_loop()
        for sig in (signal.SIGINT, signal.SIGTERM):
            try:
                loop.add_signal_handler(sig, callback=lambda: ensure_future(self._close()))
            except NotImplementedError:
                break
        for matcher in self.matchers:
            await matcher.init()
            if isinstance(matcher, EventsMatcher):
                await matcher.call(api=self.api, token=EventTypes.STARTUP)
        await self._loop()



from asyncio import run
from logging import DEBUG, INFO, WARNING, basicConfig, getLogger

from chatushka import Chatushka
from chatushka.bot.matchers.admin import admin_matcher
from chatushka.bot.matchers.users import user_matcher
from chatushka.bot.settings import get_settings

logger = getLogger()
settings = get_settings()


def make_bot(
    token: str,
    debug: bool,
) -> Chatushka:
    bot = Chatushka(token=token, debug=debug)
    bot.add_matcher(
        admin_matcher,
        user_matcher,
    )
    return bot


def cli_main(
    token: str,
    debug: bool,
) -> None:
    basicConfig(level=DEBUG if debug else INFO)
    getLogger("httpx").setLevel(WARNING)
    logger.debug("Debug mode is on".upper())
    bot = make_bot(token, debug)
    run(bot.serve())



from chatushka.core.chatushka import Chatushka
from chatushka.core.matchers.base import MatcherBase
from chatushka.core.matchers.chat_users_movements import ChatUsersMovementsEventsEnum, ChatUsersMovementsMatcher
from chatushka.core.matchers.commands import CommandsMatcher
from chatushka.core.matchers.cron import CronMatcher
from chatushka.core.matchers.events import EventsMatcher
from chatushka.core.matchers.regex import RegexMatcher
from chatushka.core.models import (
    Chat,
    ChatMemberAdministrator,
    ChatMemberOwner,
    ChatPermissions,
    EventTypes,
    MatchedToken,
    Message,
    MyChatMember,
    NewChatMember,
    OldChatMember,
    RegexMatchKwargs,
    Update,
    User,
)
from chatushka.core.protocols import MatcherProtocol
from chatushka.core.telegram import Telegram

__all__ = (
    "MatcherBase",
    "EventTypes",
    "CommandsMatcher",
    "CronMatcher",
    "EventsMatcher",
    "RegexMatcher",
    "MatcherProtocol",
    "ChatUsersMovementsMatcher",
    "ChatUsersMovementsEventsEnum",
    "Telegram",
    "User",
    "ChatMemberOwner",
    "ChatMemberAdministrator",
    "Chat",
    "NewChatMember",
    "OldChatMember",
    "Message",
    "MyChatMember",
    "Update",
    "ChatPermissions",
    "RegexMatchKwargs",
    "MatchedToken",
    "EventTypes",
    "Chatushka",
)



from random import randrange

from chatushka import CommandsMatcher, Message, RegexMatcher, Telegram
from chatushka.bot.settings import get_settings

_RESPONSE = "Зависит от контекста"

settings = get_settings()
is_four_a_lot_matcher = CommandsMatcher(
    prefixes=settings.command_prefixes,
    postfixes=settings.command_postfixes,
)
philosophy_matcher = RegexMatcher()


@philosophy_matcher(r"([/!]$)")
async def exclamation_handler(
    api: Telegram,
    message: Message,
    matched: list[str],  # pylint: disable=unused-argument
) -> None:
    rand_int = randrange(16)
    if rand_int == 1:
        await api.send_message(
            chat_id=message.chat.id,
            text=_RESPONSE,
            reply_to_message_id=message.message_id,
        )


@philosophy_matcher(r"((\s|^)(([\d]*|[а-яА-Я]*) это много/?)|((\s|^)Is ([\d]*|[a-zA-Z]*) a lot/?))")
async def philosophy_handler(
    api: Telegram,
    message: Message,
    matched: list[str],  # pylint: disable=unused-argument
) -> None:
    rand_int = randrange(4)
    if rand_int == 1:
        await api.send_message(
            chat_id=message.chat.id,
            text=_RESPONSE,
            reply_to_message_id=message.message_id,
        )



from chatushka import ChatUsersMovementsEventsEnum, ChatUsersMovementsMatcher, Message, Telegram
from chatushka.bot.settings import get_settings

settings = get_settings()
welcoming_matcher = ChatUsersMovementsMatcher()


@welcoming_matcher(ChatUsersMovementsEventsEnum.CAME, include_in_help=False)
async def came_handler(
    api: Telegram,
    message: Message,
) -> None:
    for user in message.new_chat_members:
        text = f'Привет, <a href="tg://user?id={user.id}">{user.readable_name}</a>!'
        await api.send_message(
            chat_id=message.chat.id,
            text=text,
        )



from collections.abc import Callable, Coroutine, Hashable
from datetime import UTC, datetime
from enum import Enum, auto, unique
from typing import Any, NamedTuple, Optional, TypedDict, Union

from pydantic import BaseModel, Field

# pylint: disable=invalid-name
HANDLER_TYPING = Union[
    Callable[[Any], None],
    Callable[[Any], Coroutine[Any, Any, Any]],
]


class ChatType(str, Enum):
    GROUP = "group"
    SUPERGROUP = "supergroup"
    PRIVATE = "private"


class ChatMemberStatuses(str, Enum):
    MEMBER = "member"
    CREATOR = "creator"
    ADMINISTRATOR = "administrator"
    LEFT = "left"
    KICKED = "kicked"
    RESTRICTED = "restricted"


class User(BaseModel):
    id: int
    is_bot: bool
    first_name: str | None
    last_name: str | None
    can_join_groups: bool | None
    can_read_all_group_messages: bool | None

    @property
    def readable_name(self) -> str:
        return f"{self.first_name}{' '+self.last_name if self.last_name else ''}"


class ChatMemberBase(BaseModel):
    status: ChatMemberStatuses
    user: User
    is_anonymous: bool | None = None
    custom_title: str | None = None


class ChatMemberOwner(ChatMemberBase):
    ...


class ChatMemberAdministrator(ChatMemberBase):
    can_be_edited: bool | None
    can_manage_chat: bool | None
    can_delete_messages: bool | None
    can_manage_voice_chats: bool | None
    can_restrict_members: bool | None
    can_promote_members: bool | None
    can_change_info: bool | None
    can_invite_users: bool | None
    can_post_messages: bool | None
    can_edit_messages: bool | None
    can_pin_messages: bool | None


class Chat(BaseModel):
    id: int
    type: ChatType
    title: str | None


class NewChatMember(
    ChatMemberAdministrator,
):
    ...


class OldChatMember(
    ChatMemberAdministrator,
):
    ...


class Message(BaseModel):
    message_id: int
    user: User = Field(..., alias="from")
    chat: Chat
    text: str | None
    reply_to_message: Optional["Message"] = None
    new_chat_members: list[User] = Field(default_factory=list)


Message.update_forward_refs()


class MyChatMember(BaseModel):
    chat: Chat
    user: User = Field(..., alias="from")
    date: datetime
    old_chat_member: OldChatMember
    new_chat_member: NewChatMember

    class Config:
        json_decoders = {"datetime": lambda v: datetime.fromtimestamp(v, tz=UTC)}


class Update(BaseModel):
    update_id: int
    message: Message | None = None
    my_chat_member: MyChatMember | None = None


class ChatPermissions(BaseModel):
    can_send_messages: bool
    can_send_media_messages: bool
    can_send_polls: bool
    can_send_other_messages: bool


class RegexMatchKwargs(TypedDict):
    matched: tuple[str, ...]


class MatchedToken(NamedTuple):
    token: Hashable
    args: tuple[str, ...] = ()
    kwargs: dict[str, Any] = {}


@unique
class EventTypes(Enum):
    STARTUP = auto()
    SHUTDOWN = auto()
    MESSAGE = auto()



from chatushka import CommandsMatcher
from chatushka.bot.matchers.users.bobuk_jokes import jokes_matcher
from chatushka.bot.matchers.users.eight_ball import eight_ball_matcher
from chatushka.bot.matchers.users.helpers import helpers_matcher
from chatushka.bot.matchers.users.lukashenko import lukashenko_matcher
from chatushka.bot.matchers.users.philosophy import philosophy_matcher
from chatushka.bot.matchers.users.suicide import suicide_matcher
from chatushka.bot.matchers.users.welcoming import welcoming_matcher
from chatushka.bot.settings import get_settings

settings = get_settings()
user_matcher = CommandsMatcher(
    prefixes=settings.command_prefixes,
    postfixes=settings.command_postfixes,
)
user_matcher.add_matcher(
    jokes_matcher,
    eight_ball_matcher,
    helpers_matcher,
    suicide_matcher,
    lukashenko_matcher,
    welcoming_matcher,
    philosophy_matcher,
)

__all__ = ("user_matcher",)



from click import command, option

from chatushka.bot.cli_main import cli_main


@command()
@option(
    "--token",
    "-t",
    required=True,
)
@option(
    "--debug/--no-debug",
    is_flag=True,
)
def main(
    token: str,
    debug: bool,
) -> None:
    cli_main(token=token, debug=debug)



from chatushka.matchers.base import MatcherBase
from chatushka.matchers.chat_users_movements import ChatUsersMovementsEventsEnum, ChatUsersMovementsMatcher
from chatushka.matchers.commands import CommandsMatcher
from chatushka.matchers.cron import CronMatcher
from chatushka.matchers.events import EventsMatcher
from chatushka.matchers.regex import RegexMatcher
from chatushka.models import EventTypes
from chatushka.protocols import MatcherProtocol

__all__ = (
    "MatcherBase",
    "EventTypes",
    "CommandsMatcher",
    "CronMatcher",
    "EventsMatcher",
    "RegexMatcher",
    "MatcherProtocol",
    "ChatUsersMovementsMatcher",
    "ChatUsersMovementsEventsEnum",
)
